# $Id: distributor.py,v 1.2 2003/11/13 14:51:05 hat Exp $
#
# Distributes jobs, and collects results
#
import select,sys
from general import spawn,Connection
from distributor_execprogram import Distributor_exec_program

_max_idle_delay=30    # idle exec automatically shuts down after 30 seconds

class Distributor:
    def __init__(self,forwarder,factories,machines,logobj):
	self._forwarder=forwarder  # _distributor_forwarder object
	self._factories=factories  # list of exec factories
	self._machines=machines	   # object delivering free machines
	self._busy=[]		   # list of busy exec's
	self._idle={}		   # dictionary of { hostname -> idle exec's }
	self._logobj=logobj	   # Log object
	self._joblist=[]	# list of jobs waiting for a machine

	self._uniquenum=1000	   # unique next ID number for new exec_programs

    def _collect_fds(self):
	"""Collect the file handles we are interested in.
	   return a triple (inlist,outlist,maxdelay), with
	   maxdelay==None if no timeout"""
	def _collect_collect(result,new):
	    """Both result and new are triplets (inlist,outlist,maxdelay)
	       add them together, and return the result"""
	    r0,r1,r2=result
	    for h in new[0]:
		if h not in r0:
		    r0.append(h)
	    for h in new[1]:
		if h not in r1:
		    r1.append(h)
	    if r2==None:
		r2=new[2]
	    elif new[2]==None:	# result[2]!=None and new[2]==None
		pass
	    elif r2>new[2]: # both not None, and new lower bound
		r2=new[2]
	    return (r0,r1,r2)

	result=([],[],None)
	# collect forwarder object handle
	new=self._forwarder.collect_fds()
	result=_collect_collect(result,new)
	# collect handles of idle machines
	new=self._machines.collect_fds()
	result=_collect_collect(result,new)
	# collect handles of busy exec's
	for b in self._busy:
	    new=b.collect_fds()
	    #print 'new=',new
	    result=_collect_collect(result,new)
	# collect handles of idle exec's
	for vals in self._idle.values():
	    for i in vals:
		new=i.collect_fds()
		result=_collect_collect(result,new)
	return result

    def _handle_fds(self,inl,outl):
	"""Handle IO with all our neighbours"""

	quit=self._forwarder.handle_fds(inl,outl) # handle IO with forwarder
	self._machines.handle_fds(inl,outl)	  # handle IO of machines

	# handle IO of busy execs
	newlist=[]
	for b in self._busy:
	    q=b.handle_fds(inl,outl)
	    if q:
		# an active exec died !!
		self._log('panic','Execution of job "'+str(b.read_job())
				    +'" died unexpectedly, cannot execute!!')
	    else:
		newlist.append(b)
	self._busy=newlist

	# handle IO of idle exec
	for k in self._idle.keys():
	    newlist=[]	# build a new list of idle execs for machine k
	    for i in self._idle[k]:
		q=i.handle_fds(inl,outl)
		if not q:
		    newlist.append(i)
		else:
		    self._log('info',"Idle exec at "+i.get_hostname()+" died")

	    # rebuild the self._idle[k] entry
	    if len(newlist)==0:
		del self._idle[k]
	    else:
		self._idle[k]=newlist

	return quit # 1 if the forwarder (our parent) died)


    def run(self):
	"""Run the distributor object. This is executed as child of the
	   forwarder process."""

	self._init_distributor()
	quit=0
	while 1:
	    # collect fd's + time out
	    (inlist,outlist,maxdelay)=self._collect_fds()

	    # wait for something to happen
	    if maxdelay==None:
		(inl,outl,errl)=select.select(inlist,outlist,[])
	    else:
		(inl,outl,errl)=select.select(inlist,outlist,[],maxdelay)

	    # handle IO (note that idle execs may die due to timeout here)
	    quit=self._handle_fds(inl,outl)

	    self._collect_results() # collect finished results

	    if quit:	# forwarder died, we should die too
		break

	    self._new_jobs()	# pull new jobs from the forwarder
	    self._start_jobs()	# distribute jobs over machines

        # end of loop. we should quit now
	self._closedown()
	sys.exit(0) # end of execution

    def _init_distributor(self):
	# Copy machine and jobtype information from the factories to the
	# machine objects
	for f in self._factories:
	    jtd=f.get_machines()
	    self._machines.set_machines(jtd)
	self._joblist=[]
	self._log('info','Distributor started')

    def _new_jobs(self):
	"""Move new jobs from the forwarder to the machine object"""
	j=self._forwarder.get_job()
	while j!=None:
	    if(self._machines.add_job(j.job_type(),1)==1):
		self._joblist.append(j)
		self._log('info','Job '+str(j)+' arrived')
	    else:
		self._log('serious','Job '+str(j)+' is not accepted')
	    j=self._forwarder.get_job()

    def _collect_results(self):
	newlist=[]  # construct a new list of busy execs
	for b in self._busy:
	    res=b.get_result()
	    if res == None:
		newlist.append(b)
	    else:
		self._forwarder.put_result(res)  # store result in forwarder
		j=b.read_job()	# just for outputting logging
		self._log('progress','Result of job '+str(j)+' received')
		# return machine to free machines
		host=b.get_hostname()
		self._machines.give_machine(host)
		# b must be moved to idle set
		if self._idle.has_key(host):
		    self._idle[host].append(b)
		else:
		    self._idle[host]=[b]
	self._busy=newlist

    def _start_jobs(self):
	alloc=self._machines.get_machine_jobtype()
	while alloc!=None:
	    (mach,jobtype)=alloc
	    self._log('debug','Trying to allocate type '+jobtype+' at '+mach)
	    for job in self._joblist:
		if job.job_type()==jobtype:
		    self._log('progress','Machine "'+alloc[0]+'" assigned to job '+str(job))
		    self._joblist.remove(job)
		    break
	    else:
		self._log('serious','No job with jobtype '+str(alloc[1])+' found')
		job=None
		self._machines.give_machine(mach)   # return the machine

	    # 'job' is the job to run at machine 'mach'
	    # Now find a place to run it

	    # do we have an idle exec_program to that machine?
	    if job!=None and self._idle.has_key(mach):
		for inum in range(0,len(self._idle[mach])):
		    dep=self._idle[mach][inum]
		    if dep.can_compute(job.job_type()):
			# found one, allocate the job to it
			del self._idle[mach][inum]
			dep.put_job(job)
			self._busy.append(dep)
			self._log('debug','Allocated onto an existing exec')
			job=None    # finished, do next one
			break

	    if job!=None:
		# either we didn't have idle exec's or they couldn't
		# run the job
		for f in self._factories:
		    ep=f.make_exec_program(job.job_type(),mach,self._uniquenum)
		    if ep!=None:
			# new exec_prog !
			dep=Distributor_exec_program(ep,_max_idle_delay)
			self._uniquenum=self._uniquenum+1
			dep.put_job(job)
			self._busy.append(dep)
			self._log('debug','Allocated onto a new exec')
			job=None
			break
	    if job!=None:
		# still not allocated ?
		# this is serieus !!
		self._log('panic','Execution of job "'+str(job)
						+'" at machine '+mach+' cannot be done')
		self._machines.give_machine(mach)   # return the machine

	    # find next allocation (and go back to the while alloc!=None:
	    alloc=self._machines.get_machine_jobtype()

    def _closedown(self):
	"""Close down everything"""

	for b in self._busy:
	    b.close()
	for vals in self._idle.values():
	    for i in vals:
		i.close()

	while len(self._busy)>0 or len(self._idle)>0:
	    # collect fd's + time out
	    (inlist,outlist,maxdelay)=self._collect_fds()
	    # wait for something to happen
	    if maxdelay==None:
		(inl,outl,errl)=select.select(inlist,outlist,[])
	    else:
		(inl,outl,errl)=select.select(inlist,outlist,[],maxdelay)
	    if len(inl)==0 and len(outl)==0:
		break	# quit anyway if nothing happens
	    # handle IO (note that idle execs may die due to timeout here)
	    self._handle_fds(inl,outl)



    def _log(self,etype,line):
	self._logobj.log(etype,line)
