# $Id: machines.py,v 1.2 2003/11/13 14:51:06 hat Exp $
#

class Machines:
    """Base class that performs machine management. Should be overridden with
       a derived class that actually implements a policy."""

    def __init__(self):
	pass

    def set_machines(self,jobtypedict):
	"""Setup internal machine dictionaries using the jobtype->list machines
	   dictionary 'jobtypedict' from the factory object.
	   This method is called once for each factory."""
	pass

    def add_job(self,jobtype,cnt=1):
	"""Notify that 'cnt' jobs of type 'jobtype' have been inserted (and we
	   will need machines to compute them). Returns 1 if the jobs have been
	   accepted, and 0 if not."""
	return 0    # By default, each job is refused.

    def get_machine_jobtype(self):
	"""Get a machine from the free list for creating a busy exec
	   (i.e. prepare to exec a computation).
	   Return a (hostname,jobtype) tuple or None (if nothing can be
	   selected). Also, decrement jobtype count."""
	return None

    def give_machine(self,m):
	"""Machine m is returned after use. We will keep it for future use"""
	raise "give_machine is not implemented"


    # IO stuff, for future extensions

    def collect_fds(self):
	"""Collect file handles and time out."""
	return ([],[],None) # no in-handle, no out-handle, no time-out

    def handle_fds(self,inlist,outlist):
	"""Handle IO (if our handles are in the linlist or outlist)"""
	pass

